/***************************************************************************
        engine.cpp  -  a Go playing engine based on the original wally.c
                             -------------------
    begin                : Thu Sep 19 2002
    copyright            : (C) 2002 by Andrew D. Balsa
    email                : andrebalsa@mailingaddress.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/* Function descriptions

  Main functions

    initgame()
    scanfcoo()  parses the opponent's move
    mymove()    generates the computer's move
    enemymove() gets, parses, and executes the opponent's move
    gremove()   removes a captured group from the board
    capture()   removes captured stones from board
    score()     computes the final game score using Chinese rules
    escape()
    attack()    returns the best "contact" move possible
    movedone()  complete the move after the stone is placed
    influence() calculate the influence of a stone (see documentation)

  Auxiliary functions

    rng()       random number generator
    printgame() outputs the game board
    colletter() determines the column letter from the column number
    nextp()     who plays next
    getnb()     get first non-blank character from stdin

    fatal()     these 2 routines abort the game with some kind of debug message
    panic()
 */

#include <iostream>
#include <iomanip>
#include <string>

using namespace std;

#include "game.h"
#include "engine.h"
#include "patterns.h"

// Instantiate thegame
  game thegame;


void add_influence(int x, int y, int color)  {
   int mod;       /* modifies for wally's or enemy's pieces */
   if(color == thegame.wallyColor)
     mod = 1;
   else
     mod = -1;
   influence[x][y]     += (mod*60);
   influence[x][y+1]   += (mod*13);
   influence[x][y-1]   += (mod*13);
   influence[x][y+2]   += (mod*5);
   influence[x][y-2]   += (mod*5);
   influence[x][y+3]   += (mod*1);
   influence[x][y-3]   += (mod*1);
   influence[x+1][y]   += (mod*13);
   influence[x+1][y+1] += (mod*6);
   influence[x+1][y-1] += (mod*6);
   influence[x+1][y+2] += (mod*2);
   influence[x+1][y-2] += (mod*2);
   influence[x-1][y]   += (mod*13);
   influence[x-1][y+1] += (mod*6);
   influence[x-1][y-1] += (mod*6);
   influence[x-1][y+2] += (mod*2);
   influence[x-1][y-2] += (mod*2);
   influence[x+2][y]   += (mod*5);
   influence[x+2][y+1] += (mod*1);
   influence[x+2][y-1] += (mod*1);
   influence[x+2][y+2] += (mod*1);
   influence[x+2][y-2] += (mod*1);
   influence[x-2][y]   += (mod*5);
   influence[x-2][y+1] += (mod*1);
   influence[x-2][y-1] += (mod*1);
   influence[x-2][y+2] += (mod*1);
   influence[x-2][y-2] += (mod*1);
   influence[x+3][y]   += (mod*1);
   influence[x-3][y]   += (mod*1);
}

/***************************************************************************
 *                                                                         *
 *                        Auxiliary Functions                              *
 *                                                                         *
 ***************************************************************************
 */


void fatal(char *error_message) {

    /*Print an error message and abort program.*/

    cerr << "\n?!fatal error--" << error_message << endl;
    exit(EXIT_FAILURE);
}


void panic(char *panic_message) {

    /*Fatal error routine reserved for error conditions that ought not happen.*/

    cerr << "\n?!panic--" << panic_message << endl;
    exit(EXIT_FAILURE);
}


int nextp (int p) {

/*Return the code for the other player, who is the next to play.*/

int np = BLACK;

if (p == BLACK) np = WHITE;
else if (p == WHITE) np = BLACK;
     else panic("illegal input to nextp()");
return np;
}


int rng(int n) {

    /*Return a random number from 0 to n-1.*/
    /*(This keeps the program's moves from all lying in a trivial pattern.)*/

    int r;
    r = rand() % n;
    return r;
}

void printgame()
/*
Output the game board theboard (with ko info from thegame) to the console.
Currently the style is:

   a b c d e f g h j
 1 . . . . . . . . . 1
 2 . . . . . O O . . 2
 3 . . O . . . # . . 3
 4 . . O . . . # . . 4
 5 O O ^ O . . . . . 5
 6 . # O # . # # . . 6
 7 . # # # . . . . . 7
 8 . . . . . . . . . 8
 9 . . . . . . . . . 9
   a b c d e f g h j

where '^' is a ko, '#' is black, 'O' is white.
*/
{ int x, y;

  cout << endl << "  ";
  for(x=0; x<thegame.boardSize; ++x) {
    cout << " ";
    cout.put(lettercol(x));
  }
  cout << "\n";
  for(y=0; y<thegame.boardSize; ++y)
  { cout.setf( ios::right, ios::adjustfield );
    cout.width(2);
    cout << (1+y) << " ";
    for(x=0; x<thegame.boardSize; ++x)
    { register int here= theboard[x][y];
      printf( x!=(thegame.boardSize-1)?"%c ":"%c",
		(thegame.kox==x&&thegame.koy==y) ? '^' :
		(EMPTY==here) ? '.' :
		(WHITE==here) ? 'O' :
		(BLACK==here) ? '#' :
//		Replaced
//		panic("illegal board type in printboard()"));
//		by
		'@');	// call the panic function later
    }
    cout << " " << flush;
    cout.unsetf( ios::right );
    cout.width(2);
    cout.setf( ios::left, ios::adjustfield );
    cout << (y+1);
    cout.unsetf( ios::left );
    if ( y == 0 )
      cout << "     BLACK has captured " << thegame.capturedWhite << " stones.";
    if ( y == 1 )
      cout << "     WHITE has captured " << thegame.capturedBlack << " stones.";
    cout << endl;
  }
  cout << flush;
  cout << "  ";
  for(x=0; x<thegame.boardSize; ++x) {
    cout << " ";
    cout.put(lettercol(x));
  }
  cout << endl;

#ifdef TEST_INFLUENCE
  int idx1, idx2;
  for(idx1 = 0; idx1 < thegame.boardSize; idx1++)   {
      for(idx2 = 0; idx2 < thegame.boardSize; idx2++)   {
        cout.width(3);
        cout << influence[idx2][idx1];
      }
      cout << endl;
   }
#endif /*TEST_INFLUENCE*/

}


void initgame() {

  register int x, y;

/* Initialize the Random Number Generator */
  srand(RANDOM_SEED);
  
/*Clear the board.*/
  for(x=0; x<thegame.boardSize; ++x)  for(y=0; y<thegame.boardSize; ++y)
    theboard[x][y]= EMPTY;

/*Initialize miscellaneous other stuff.*/
  thegame.kox= thegame.koy= (-1);
  thegame.capturedWhite = thegame.capturedBlack = 0;
  thegame.tur= 1;
  thegame.qpa= 0;
  lex = ley = (-1000);	/*nothing is near the enemy's last move*/

// Place the fixed handicap stones
  place_fixed_handicap();
}

void place_fixed_handicap()
{
  // place the stones based on board size and build
  // the handicap stones vertices list simultaneously

  // clear the handicap stones vertices list first
  handicapList.erase();

  if(thegame.blackHandicap > 1){
    if (thegame.boardSize == 9)
      setHandicap9(thegame.blackHandicap);
    if (thegame.boardSize == 13)
      setHandicap13(thegame.blackHandicap);
    if (thegame.boardSize == 19)
      setHandicap19(thegame.blackHandicap);
  }
}
void place_handicap_stone (int x, int y) {
  bnassert(onboard(x,y));
  bnassert(EMPTY==theboard[x][y]);
  theboard[x][y]= BLACK;
  add_influence(x, y, BLACK);
  // add vertex to handicapList
  handicapList += (char)lettercol(x);
  if ( y >= 9 ){
    handicapList += "1";
    if ( y == 9 )
      handicapList += "0";
    else
      handicapList += ('1' + ( y % 10 ));
  }
  else
    handicapList += ( '1' + y );
  handicapList += " ";
}

// Handicap stone placement code adapted from program AmiGo by Todd R. Johnson
void setHandicap9(int handicap9) {
 if (handicap9 < 2)
    return;

  place_handicap_stone(2, 2);
  place_handicap_stone(6, 6);

  if (handicap9 >= 3)
    place_handicap_stone(2, 6);
  if (handicap9 >= 4)
    place_handicap_stone(6, 2);
}

void setHandicap13(int handicap13) {
 if (handicap13 < 2)
    return;

  place_handicap_stone(3, 3);
  place_handicap_stone(9, 9);

  if (handicap13 >= 3)
    place_handicap_stone(3, 9);
  if (handicap13 >= 4)
    place_handicap_stone(9, 3);
  if (handicap13 == 5 || handicap13 == 7 || handicap13 == 9)
    place_handicap_stone(6, 6);
  if (handicap13 >= 6) {
    place_handicap_stone(9, 6);
    place_handicap_stone(3, 6);
  }
  if (handicap13 >= 8) {
    place_handicap_stone(6, 9);
    place_handicap_stone(6, 3);
  }
}

void setHandicap19(int handicap19) {
 if (handicap19 < 2)
    return;

  place_handicap_stone(3, 3);     // D4
  place_handicap_stone(15, 15);   // Q16

  if (handicap19 >= 3)
    place_handicap_stone(3, 15);  // D16
  if (handicap19 >= 4)
    place_handicap_stone(15, 3);  // Q4
  if (handicap19 == 5 || handicap19 == 7 || handicap19 == 9)
    place_handicap_stone(9, 9);   // K10
  if (handicap19 >= 6) {
    place_handicap_stone(3, 9);   // D10
    place_handicap_stone(15, 9);  // Q10
  }
  if (handicap19 >= 8) {
    place_handicap_stone(9, 3);   // K4
    place_handicap_stone(9, 15);  // K16
  }
}

/***************************************************************************
 *                                                                         *
 *                            Main Functions                               *
 *                                                                         *
 ***************************************************************************
 */


void gremove(int x, int y)

/*
 Recursively remove the stone at (x,y) and all stones in the same group
 from the board.
*/
{ register int color;
  bnassert(onboard(x,y));
  color= theboard[x][y];
  bnassert(BLACK==color||WHITE==color);
  theboard[x][y]= EMPTY;

  if(onboard(x, y+1) && color==theboard[x][y+1])
    gremove(x, y+1);
  if(onboard(x, y-1) && color==theboard[x][y-1])
    gremove(x, y-1);
  if(onboard(x+1, y) && color==theboard[x+1][y])
    gremove(x+1, y);
  if(onboard(x-1, y) && color==theboard[x-1][y])
    gremove(x-1, y);
}


int capture(int p, int *capx, int *capy)
/*
Remove all of player p's dead stones (as implied by results of makegroups())
from the board.  Return # stones removed.  If any stones are removed,
theboard[][] is left inconsistent with groups[].
*capx and *capy are set to the coordinates of a captured stone, if any,
so that if only one stone is captured, they may be used to set ko.
*/
{ register int j, rv=0;
  register group *g;
  for(g=groups,j=0; j<ngroups; ++g,++j)
    if(p==g->color&&0==g->nliberties)
    { rv += g->nstones;
      *capx= g->x;
      *capy= g->y;
      gremove(g->x, g->y);
    }
  return rv;
}


int colletter(int letter)
/*
Return the column # (0 to boardsize-1) which corresponds to letter.
Due to perversity of tournament organizers etc., 'i' or 'I' is
omitted, leading to an increase in complexity of this routine.
Return (-1) on illegal input.
*/
{ register int result;

  if('a'<=letter&&letter<='z')
    result= letter-'a';
  else if('A'<=letter&&letter<='Z')
    result= letter-'A';
  else
    return (-1);

  if(8>result)
    ;
  else if(8==result)
    return (-1);
  else
    --result;

  if(result >= thegame.boardSize)
    return (-1);
  else
    return result;
}


int getnb()
/*Read and return the first nonblank character from the standard input.*/
{ int c;
  do
    c= getchar();
  while(' '==c||'\n'==c||TAB==c);
  return c;
}


int scanfcoo(int *x, int *y)
/*
Read a coordinate from the standard input and write it to *x and *y.  Like
scanf, return 1 for success, 0 for failure, negative numbers for special
moves, e.g. PASS and RESIGN.
On any input error, the function eats input up to the next newline.
*/
{
  int xx, yy, c1, c2;

/*Get first nonblank character of input, normally a letter code for a column.*/
  c1= getnb();

/*Check for possible resignation (a '.' at the start of the input).*/
  if('.'==c1)
  { while('\n'!=getchar())	/*Skip to end of line.*/
      ;
    return RESIGN;
  }

/*Check for possible "pass".*/
  if('p'==lowercase(c1))
  { c2= getchar();
    if('a'!=lowercase(c2))
      ungetc(c2, stdin);
    else
    { if((c2=getchar(),'s'!=lowercase(c2))||(c2=getchar(),'s'!=lowercase(c2)))
      { ungetc(c2, stdin);
	goto skipquit;
      }
      else
        return PASS;
  } }
  xx= colletter(c1);
  if(1!=scanf("%d", &yy))
  {
skipquit:	/*Skip to the next newline and return failure.*/
    while('\n'!=getchar())
      ;
    return 0;
  }
  if(0==onboard(xx,yy-1))	/*The -1 converts between C arrays starting */
				/* at 0 and human boards counting from 1.*/
    goto skipquit;
  else
  { *x= xx;
    *y= yy-1;
    return 1;
  }
}


void movedone()
/*
Do everything which must be done to complete a move after the stone is
placed.
*/
{ thegame.pla= nextp(thegame.pla);  ++thegame.tur; }


void lcount(int x, int y, group *thisgroup, board scratch, int mark)

/*
Recursively group together connected stones.  Three things must be done:
  (1) count liberties in thisgroup->nliberties,
  (2) count stones in thisgroup->nstones, and
We set scratch[][]= mark for each point and liberty of this group,
  so that we only count each only once.  The calling routine must
  see to it that scratch[][]!=mark for all points and liberties that
  it wants counted.
*/
{ register int *bxy, *sxy; /*Common subexpressions to reduce array */
  		/* arithmetic; they point to theboard[x][y] and scratch[x][y].*/

endrecurse:
  bnassert(onboard(x,y));
  bnassert(thisgroup->color==theboard[x][y]);
  bnassert(thisgroup->nstones>=0);
  bnassert(thisgroup->nliberties>=0);
  bnassert(mark!=scratch[x][y]);

	/*Set common subexpressions.*/
  bxy= &(theboard[x][y]);
  sxy= &(scratch[x][y]);

	/*Count the point (x,y) and make it a member of the group.*/
  ++(thisgroup->nstones);
  *sxy= mark;

	/*"Process" (x, y-1): should it be in the group, or a liberty, or */
	/* nothing?*/
  y= y-1;
  bxy -= 1;
  sxy -= 1;
  bnassert(&theboard[x][y]==bxy);
  bnassert(&scratch[x][y]==sxy);
  if(y>=0)
    if(thisgroup->color==*bxy)
    { if(*sxy!=mark)
        lcount(x, y, thisgroup, scratch, mark);
    }
    else if(EMPTY==*bxy)
    { if(*sxy!=mark)
      { *sxy=mark;
	++(thisgroup->nliberties);
    } }

	/*Process (x, y+1).*/
  y= y+2;
  bxy += 2;
  sxy += 2;
  bnassert(&theboard[x][y]==bxy);
  bnassert(&scratch[x][y]==sxy);
  if(y<thegame.boardSize)
    if(thisgroup->color==*bxy)
    { if(*sxy!=mark)
        lcount(x, y, thisgroup, scratch, mark);
    }
    else if(EMPTY==*bxy)
    { if(*sxy!=mark)
      { *sxy=mark;
	++(thisgroup->nliberties);
    } }

	/*Process (x-1, y).*/
  y= y-1;
  x= x-1;
  bxy -= EDGE+1;
  sxy -= EDGE+1;
  bnassert(&theboard[x][y]==bxy);
  bnassert(&scratch[x][y]==sxy);
  if(x>=0)
    if(thisgroup->color==*bxy)
    { if(*sxy!=mark)
        lcount(x, y, thisgroup, scratch, mark);
    }
    else if(EMPTY==*bxy)
    { if(*sxy!=mark)
      { *sxy=mark;
	++(thisgroup->nliberties);
    } }

	/*Process (x+1, y).*/
  x= x+2;
  bxy += 2*EDGE;
  sxy += 2*EDGE;
  bnassert(&theboard[x][y]==bxy);
  bnassert(&scratch[x][y]==sxy);
  if(x<thegame.boardSize)
    if(thisgroup->color==*bxy)
    { if(*sxy!=mark)
        goto endrecurse;
    }
    else if(EMPTY==*bxy)
    { if(*sxy!=mark)
      { *sxy=mark;
	++(thisgroup->nliberties);
    } }

}


void makegroups()
/*Analyzes groupings of stones on board; sets groups[].*/
{
  register int x, y;	/*coords of point we are now assigning to a group*/
  board scratch;	/*scratch for lcount() to use to say if a point is */
			/* already counted*/

	/*Clear old groups[] table and scratch[][].*/
  ngroups= 0;
  { register int *s;
    for(s=scratch[0],x=0; x<EDGE*EDGE; ++x)
      *s++= 0;
  }

	/*Go through board[][] point by point.  Group any stone which has not */
	/* already been grouped (as indicated by negative scratch[][]).*/
  { register int *bxy= (int*)theboard, *sxy= (int*)scratch;
    register group *thisgroup=groups;
    for(x=0; x<thegame.boardSize; ++x,bxy+=EDGE-thegame.boardSize,sxy+=EDGE-thegame.boardSize)
      for(y=0; y<thegame.boardSize; ++y,++bxy,++sxy)
      { bnassert(&(theboard[x][y])==bxy);
	bnassert(&(scratch[x][y])==sxy);
	bnassert(&(groups[ngroups])==thisgroup);
        if((BLACK==*bxy||WHITE==*bxy) && 0==*sxy)
        { thisgroup->color= *bxy;
          thisgroup->x= x;
          thisgroup->y= y;
          thisgroup->nliberties= 0;
          thisgroup->nstones= 0;
          lcount(x, y, thisgroup, scratch, 1+thisgroup-groups);
          ++thisgroup;
          ++ngroups;
        }
      }
  }
}


void placestone(int x, int y, int p)
/*
 * Put a stone for player p at (x,y) and remove all the captures which result.
 * Update ko information.
 */

{
  int ncap, capx, capy;
  bnassert(onboard(x, y));
  bnassert(BLACK==p||WHITE==p);
  bnassert(EMPTY==theboard[x][y]);

  theboard[x][y]= p;
  add_influence(x, y, p); /* Use new add_influence function */
  makegroups();           /* Group all stones and count libs.*/
  ncap = capture(nextp(p), &capx, &capy); /* Remove p's opponent's dead stones.*/

  // update captured stones count for player p
  if ( p == BLACK )
    thegame.capturedWhite += ncap;
  else
    thegame.capturedBlack += ncap;
  thegame.kox= thegame.koy= (-1);         /* Cancel any old ko.*/

  if(1==ncap)             /*If exactly one stone was captured, check for new ko.*/
      { register int j, *ip;
        group g;
        board scratch;

        for(ip=(int*)scratch,j=EDGE*EDGE; j--; ) *ip++= 0;
        g.color= theboard[x][y];
        g.nstones= 0;
        g.nliberties= 0;
        g.x= x;
        g.y= y;
        lcount(x, y, &g, scratch, 1); /* Count stones and libs for capturing group.*/
        if(1==g.nstones && 1==g.nliberties) {	/*If c.g. is a single stone in atari */
	      thegame.kox= capx;
	      thegame.koy= capy;
	} /* then ko, so mark it.*/
  }

  // (ADB) I have some trouble understanding the following lines
  // review them later:
  //     - should I update the captured stones counts or not ?
  //     - does wallyplus really know not to make illegal suicide moves ?

  if(ncap)		    // If any enemy stones were removed,
    makegroups(); // recalculate the board.
  ncap= capture(p, &capx, &capy);     // Remove any still-trapped stones (suicide).
  if ( thegame.wallyColor == BLACK )
    bnassert( WHITE==p || ncap==0 );  // Wally knows not to make suicides !
  else
    bnassert( BLACK==p || ncap==0 );
}



void sortv(group **p, int n)
/*
Sort the table p[] or pointers to groups in order of vulnerability,
most vulnerable first.
Currently this is the slow simple bubble sort we all know.
*/
{ register int uns;	/*flag that an unsorted pair was found last pass*/
  register group *t,	/*exchange temporary*/
		 **ip;	/*pointer through p[]*/
  register int j;	/*loop index through p*/

  do
  { uns= 0;
    for(j=0,ip=p; j<n-1; ++j,++ip)
    { t= ip[1];
      if( t->nliberties< (*ip)->nliberties ||
	 (t->nliberties==(*ip)->nliberties && t->nstones>(*ip)->nstones) )
      { uns= 1;
        ip[1]= *ip;
	*ip= t;
      }
    }
  } while(uns);

}


int subj_lib(int x, int y, int color)
/*
 Return the resulting liberties of the group created by a <color> move at x, y,
 that is, how many liberties would it have after the move was made?

 This is done shoddily, without considering the liberties created by
 the disappearance of any groups captured by the move.
*/
{ board scratch;	/*parameter array of flags for lcount()*/
  group t;		/*parameter group for lcount()*/

  bnassert(onboard(x,y));
  bnassert(EMPTY==theboard[x][y]); /*Make sure we can undo by restoring EMPTY.*/

  theboard[x][y]= color;	/*Temporarily place the stone in question.*/

/*Clear scratch[][] so lcount() works properly.*/
  { register int *s, j;
    for(j=EDGE*EDGE,s=(int*)scratch; j--; )
      *s++= 0;
  }

/*Make a fake group for the new stone to be part of -- this is */
/* a necessary parameter for lcount().*/
  t.color= color;
  t.x= x;
  t.y= y;
  t.nliberties= 0;
  t.nstones= 0;

/*Show this little Potemkin village to the lcount().*/
  lcount(x, y, &t, scratch, 1);

/*Undo the temporary move (x,y).*/
  theboard[x][y]= EMPTY;

/*Return the result from lcount().*/
  return t.nliberties;

}


void pattern1(int *u, board masks, board movehere)

/*
A routine called by pattern() to match possible moves to
tabulated patterns of specific orientation.
*u gives the urgency of the best move found so far; we
are to ignore moves which are inferior to this.

If this move has not already been found (compare to the
urgency value in movehere[]) then we set its urgency
value in *u and in movehere[], and put it into the table
goodmoves[].  If the move is better than any found before,
we put it at the beginning of the table; otherwise we
put it in at the pointer pgoodmoves.

masks[][] is an array which represents the pieces on the board,
translated into bit masks by lookup in flookup[].

A late-breaking modification is that only the EMPTY points which have
movehere[][] set are tested, so that pattern() can be called to look for
contact plays or possibly other restricted sets which make good patterns.
An even later modification permits the use of movehere[][] to
tell the urgency of the best move which has been made to that
point.
*/
{ register int *is, *iis;/*pointers into patterns[]; or scratch*/
  register int j;	/*& into a particular pattern, # points remaining*/
  register int x, y;	/*current position we're trying to match pattern to*/
  register int xs, ys;	/*position of a point from a pattern*/
  int ua;		/*urgency and adjustment for this move*/
  int thispat;		/*which pattern in the table are we currently */
			/* trying to match?*/


  for(is=patterns,thispat=0; PATTERN==*is; is+= 3+3*(is[1]),++thispat)
  { for(x=0; x<thegame.boardSize; ++x)
      for(y=0; y<thegame.boardSize; ++y)
      { if(F_EMPTY==masks[x][y] && 0==ko(x,y)
         &&(thegame.kox!=x||thegame.koy!=y) )
        { for(iis=is+1,j= *iis++,ua= *iis++; j; --j)
	  { xs= *iis++;
	    ys= *iis++;
	    if(onboard(x+xs,y+ys))
	    { if(0==(masks[x+xs][y+ys]&*iis++)) goto mismatch; }
	    else
	    { if(0==(F_OFF&*iis++)) goto mismatch; }
	  }
	  /*If we fall through here, then we matched a pattern.*/
	  /*Compute adjusted urgency.*/
	  if(abs(x-lex)<=1&&abs(y-ley)<=1)
	    ua-=4;	            /*Important to oppose enemy's last move.*/
	  for(xs=(-2); xs<=2; ++xs)
	    for(ys=(-2); ys<=2; ++ys)
	      if(onboard(x+xs,y+ys))/*Important to move to */
	        ++ua;		    /* uncrowded parts of the board.*/
	  if( ua<*u &&		    /*If this pattern is most urgent so far */
	      0==intoatari(x, y) )  /* and the move is not futile */
	  { *u= ua;	 	    /*Replace old */
	    movehere[x][y]= ua+(2*influence[x][y]);     /* urgency values.*/
	    patnum= thispat;	    /*Record pattern # for debugging.*/
	    pgoodmoves= goodmoves;  /*Reinit goodmoves[].*/
	    goto intogoodmoves;
	  }
	  else if( ua==*u &&	    /*If there's no better pattern */
	           ua<movehere[x][y] /* and it's the best move here */
		   && !intoatari(x,y)) /*and not futile*/
	  { movehere[x][y]= ua;	/*Mark as best move here.*/
	  intogoodmoves: ;	/*Put it into goodmoves[].*/
	    *pgoodmoves++= x;  *pgoodmoves++= y;
	  }
        }
	mismatch: ;
      }
  }
  if(PATTEND!=*is)
  { cerr << "?error in pattern table, pattern #" << thispat << endl;
    panic("programmer error in pattern table");
  }
}


void pattern(int *chosenx, int *choseny, int *urgency, board movehere)

//void pattern(int chosenx, int choseny, int urgency, board movehere)
// int *chosenx, *choseny, *urgency;  board movehere;

/*
   Try to find a good black(?) move by matching to a table of patterns.
   Returns the urgency of the move; returns BIGU if no match found.

   In order to match the patterns in all orientations, we
   reflect the entire table eight times, checking for a match each time.
   The reflections are
    (1)   y <-> thegame.boardSize-1-y	(across a mirror plane parallel to x-axis)
    (2)	x <-> thegame.boardSize-1-x	( " y-axis)
    (3)   y <-> thegame.boardSize-1-y	( " x-axis)
    (4)   x <-> y		( " the line y=x)
    (5)   y <-> thegame.boardSize-1-y	(across a mirror plane parallel to x-axis)
    (6)	x <-> thegame.boardSize-1-x	( " y-axis)
    (7)   y <-> thegame.boardSize-1-y	( " x-axis)
    (8)   x <-> -y	( " the line y=(-x))
   Draw pictures if necessary to see how this works.
*/
{
  register int x, y, t, j, *is;
  board scratch;


/*Translate the board to flags for easy comparison with pattern table.*/
  for(x=0; x<thegame.boardSize; ++x)
    for(y=0; y<thegame.boardSize; ++y)
      scratch[x][y]= flookup[theboard[x][y]];

  *urgency= BIGU;		/*No big move so far.*/
  pgoodmoves= goodmoves;	/*No moves so far.*/
  patnum= (-1);			/*For debugging: no pattern # so far.*/

  pattern1(urgency, scratch, movehere);
				/*Find matches to untransformed table.*/

/*Invert y coordinates in pattern table.*/
  for(is=patterns; PATTERN==*is; )
  { for(j= *++is,is+=2; j--; )
    { is += 1;
      *is= (-*is);
      is += 2;
    }
  }
  bnassert(PATTEND==*is);


  pattern1(urgency, scratch, movehere);
				/*Find matches to transformed table.*/

/*Invert x coordinates in pattern table.*/
  for(is=patterns; PATTERN==*is; )
  { for(j= *++is,is+=2; j--; )
    { ;
      *is= (-*is);
      is += 3;
    }
  }
  bnassert(PATTEND==*is);

  pattern1(urgency, scratch, movehere);
				/*Find matches to transformed table.*/

/*Invert y coordinates in pattern table.*/
  for(is=patterns; PATTERN==*is; )
  { for(j= *++is,is+=2; j--; )
    { is += 1;
      *is= (-*is);
      is += 2;
    }
  }
  bnassert(PATTEND==*is);

  pattern1(urgency, scratch, movehere);
				/*Find matches to transformed table.*/

/*Exchange x and y coordinates in pattern table.*/
  for(is=patterns; PATTERN==*is; )
  { for(j= *++is,is+=2; j--; )
    { t= *is;
      *is= is[1];
      is[1]= t;
      is += 3;
    }
  }
  bnassert(PATTEND==*is);

  pattern1(urgency, scratch, movehere);
				/*Find matches to transformed table.*/

/*Invert y coordinates in pattern table.*/
  for(is=patterns; PATTERN==*is; )
  { for(j= *++is,is+=2; j--; )
    { is += 1;
      *is= (-*is);
      is += 2;
    }
  }
  bnassert(PATTEND==*is);

  pattern1(urgency, scratch, movehere);
				/*Find matches to transformed table.*/

/*Invert x coordinates in pattern table.*/
  for(is=patterns; PATTERN==*is; )
  { for(j= *++is,is+=2; j--; )
    { ;
      *is= (-*is);
      is += 3;
    }
  }
  bnassert(PATTEND==*is);

  pattern1(urgency, scratch, movehere);
				/*Find matches to transformed table.*/

/*Invert y coordinates in pattern table.*/
  for(is=patterns; PATTERN==*is; )
  { for(j= *++is,is+=2; j--; )
    { is += 1;
      *is= (-*is);
      is += 2;
    }
  }
  bnassert(PATTEND==*is);

  pattern1(urgency, scratch, movehere);
				/*Find matches to transformed table.*/

/*Exchange x and -y coordinates in pattern table.*/
  for(is=patterns; PATTERN==*is; )
  { for(j= *++is,is+=2; j--; )
    { t= (-*is);
      *is= (-is[1]);
      is[1]= t;
      is += 3;
    }
  }
  bnassert(PATTEND==*is);

/*Try to pick a move from the table of equally good ones.*/
  { int nmoves= (pgoodmoves-goodmoves)/2;
    if(0==nmoves) return;
    else
    { int whichmove= rng(nmoves);
      *chosenx= goodmoves[2*whichmove];
      *choseny= goodmoves[2*whichmove+1];
    }
  }

/*We'd better not have decided on an illegal move.*/
#if ASSERT
  if( *urgency<BIGU )
    if( EMPTY!=theboard[*chosenx][*choseny] || ko(*chosenx,*choseny) )
      panic("illegal move selected in pattern()");
#endif

}


int attack(group *g, int *rx, int *ry, int ml)

/*
Set *rx and *ry to a move which attacks group *g and still has
at least ml liberties.

This routine tries to return the "best" contact play if several are possible.
This is currently implemented as the one as close to the
 (2.9)-th line as possible, that is, one on the third line is best,
 then one on the fourth, then one on the second, then one on the fifth,
 then etc.

Return TRUE on failure.
*/
{ int bestx, besty;        /* best contact move found so far */
  bestx = besty = -10000;  /* not on board */
  int edist;               /* 10*distance of this play from (2.9)-st line*/
  register int x, y;

  board scratch;

/*Clear scratch[][] so that lcount() can work.*/
  { register int j, *is;
    for(j=EDGE*EDGE,is=(int*)scratch;  j--; )
      *is++= 0;
  }

/*Set scratch[][] again for the stones of the group, and recalculate the */
/* number of liberties of the group.  Also set scratch[][] for all the EMPTY */
/* points which are contact moves.*/
  g->nstones= 0;	/*Clear these before */
  g->nliberties= 0;	/* passing to lcount(), which increments them.*/
  lcount(g->x, g->y, g, scratch, 1);

/*Print out scratch[][] as 9x9 table.*/
  if(debugattack) {
    printf("$In attack after lcount(), board and scratch are:\n");
    for(y=thegame.boardSize-1; y>=0; --y)
    { printf("$");
      for(x=0; x<thegame.boardSize; ++x)
      { register int c, t=theboard[x][y];
        if(EMPTY==t) c= '.';
        else if(BLACK==t) c= '#';
        else if(WHITE==t) c= 'O';
        else c= '?';
        printf("%c%d ", c, scratch[x][y]);
      }
      printf("\n");
    }
  }

/*Check that there are enough liberties to find a contact play.*/
  bnassert(g->nliberties>=0);

/*Look for the best adjacent point.*/
  edist= thegame.boardSize*10;	/*We haven't found a point near the (2.9)-th line yet.*/
  for(x=0; x<thegame.boardSize; ++x)  			/*For */
    for(y=0; y<thegame.boardSize; ++y)			/* all */
      if ( EMPTY == theboard[x][y] && scratch[x][y]	/*  contact points*/
       && (thegame.kox!=x||thegame.koy!=y) )	/*   which don't violate ko */
        if ( subj_lib( x, y, thegame.wallyColor ) >= ml )// if not too suicidal
	{ register int toedge= x;	/*distance to edge of board*/
	  if(toedge>thegame.boardSize-1-x) toedge= thegame.boardSize-1-x;
          if(toedge>y) toedge= y;
          if(toedge>thegame.boardSize-1-y) toedge= thegame.boardSize-1-y;
          if(debugattack)
            fprintf(stderr, "$Looking at %d,%d, toedge=%d, abs=%d.\n",
		x, y, toedge, abs(10*toedge-21));
	  if(abs(10*toedge-19)<edist)	/*If closer to 2.9th line than before*/
          { edist= abs(10*toedge-21);	/* record */
            if(debugattack) fprintf(stderr, "$Replacing edist.\n");
	    bestx= x;  besty= y;  /*  as the new best so far.*/
          }
          else
	    if(debugattack) fprintf(stderr, "$Not replacing edist=%d.\n", edist);
        }

  if(debugattack) fprintf(stderr, "$Edist= %d.\n", edist);

/*Return best result, if any.*/
  if(edist<thegame.boardSize*10)			/*If some good contact play found */
  { *rx= bestx;  *ry= besty;  		/* return it */
    if(debugattack) fprintf(stderr, "$Returning 0.\n");
    return 0;				/*  and success */
  }
  else
  { if(debugattack) fprintf(stderr, "$Returning 1.\n");
    return 1;
} }


int escape(group *g, int *rx, int *ry)

/*
Set *rx and *ry to a play which lets group *g escape from atari.
*/

{ register int x, y;

  board scratch;

/*Clear scratch[][] so that lcount() can work.*/
  { register int j, *is;
    for(j=EDGE*EDGE,is=(int*)scratch;  j--; )
      *is++= 0;
  }

/*Set scratch[][] again for the stones of the group, and recalculate the */
/* number of liberties of the group.  Also set scratch[][] for all the EMPTY */
/* points which are contact moves.*/
  g->nstones= 0;	/*Clear these before */
  g->nliberties= 0;	/* passing to lcount(), which increments them.*/
  lcount(g->x, g->y, g, scratch, 1);

/*Check that there are enough liberties to find a contact play.*/
  bnassert(g->nliberties>=0);

/*Return any point which succeeds.*/
  for(x=0; x<thegame.boardSize; ++x)  			/*For */
    for(y=0; y<thegame.boardSize; ++y)			/* each */
      if(EMPTY==theboard[x][y] && scratch[x][y] /*  contact point */
       &&(thegame.kox!=x&&thegame.koy!=y) )	/*   which is not ko */
        if(0==intoatari(x,y))			/*    if not into atari */
	{ *rx= x;  *ry= y;  return 0; }		/*     then return it.*/

/*Otherwise return failure.*/
  return 1;
}


int enginemove()
/*Calculate and execute the engine's move.*/
{
  register int j;  register group *g; /*index and pointer through groups[]*/
  int x, y;			/*coords of the move selected*/
  group *gt[EDGE*EDGE];		/*table of &groups in order of vulnerability*/
  group **fs, **es;		/*pointers into gt: start of friendly and */
				/* enemy tables*/
  register group **igt;		/*scratch pointer through gt[]*/
  int nf, ne;			/*number of enemy and friendly groups*/
  int upattern, patternx, patterny;
  board scratch;

/* Find the most urgent move to improve our shape.*/
/* First say that all moves are to be considered.*/
  { register int *ip;
    for(ip=(int*)scratch,j=EDGE*EDGE; j--; )
      *ip++= BIGU;
  }
/* Then call pattern search.*/
  pattern(&patternx, &patterny, &upattern, scratch);

/* Make tables of friendly and enemy groups in gt[] at *fs and *es.*/
  for(j=0,g=groups,nf=0,fs=igt=gt; j<ngroups; ++j,++g)
    if(thegame.wallyColor==g->color)
    { ++nf;
      *igt++= g;
    }
  for(j=0,g=groups,ne=0,es=igt; j<ngroups; ++j,++g)
    if( nextp(thegame.wallyColor) == g->color )
    { ++ne;
      *igt++= g;
    }

/* Sort the tables in order of vulnerability to attack.*/
  sortv(fs, nf);
  sortv(es, ne);

/* Consider extremely urgent pattern moves.*/
  if(upattern<16)
  {
  movepattern:
    if(debugattack)
      printf("$making pattern, pattern #=%d, urgency=%d.\n", patnum, upattern);
    x= patternx;
    y= patterny;
    goto movexy;
  }
/* Capture the enemy!*/
  while(ne && 1==(*es)->nliberties)
  { if(0==attack(*es, &x, &y, 0))
    { if(debugattack) printf("$Capturing the enemy.\n");
      goto movexy;
    }
    else /*(This case comes up if we can't capture because of ko.)*/
    { ++es; --ne; }		/*Go on to next most vulnerable group.*/
  }
  if(upattern<24) goto movepattern;
/*If we can't do that, then try to protect ourself.  But as per Dr. Millen, */
/* don't risk everything crawling along the edge without lookahead.*/
/*However, we have a more powerful computer than the good doctor, so we */
/* can afford to check another condition: if crawling connects to a healthy */
/* group immediately (as when the program makes a pattern match with */
/*	O O #	*/
/*	. . O .	*/
/*	~ ~ ~ ~ */
/* where '~' is off the board, then is attacked with */
/*	O O #   */
/*	. . O # */
/*	~ ~ ~ + */
/* it seems to make no sense to prohibit the program from defending with */
/*	O O #   */
/*	. O O # */
/*	~ ~ ~ ~ */
/* ..  so we *don't* prohibit it.*/
  while(nf && 1==(*fs)->nliberties)
  { if(escape(*fs, &x, &y))	/*If we can't see how to escape */
    { ++fs; --nf; }		/* go on to next friendly group, */
    else			/*  otherwise */
    { if((0==x||thegame.boardSize-1==x||0==y||thegame.boardSize-1==y) && /* if crawling */
	 (2 >= subj_lib( x, y, thegame.wallyColor )) ) // and not connecting to strong group
      { ++fs; --nf;  continue; } /* this is probably not a good move */
      else			/* but if not crawling or connecting to */
				/* a strong group, go for it.*/
      if(debugattack)
        cout << "$Protecting a friendly group from imminent capture." << endl;
      goto movexy;
    }
  }
  if(upattern<34) goto movepattern;
/*Try putting the enemy in atari.*/
  while(ne && 2==(*es)->nliberties)	/*For all vulnerable enemy groups */
  { if(attack(*es, &x, &y, 2))		/* if we can't see how to attack */
    { ne--; ++es; }			/*  go on to the next enemy group, */
    else				/*   otherwise */
    { if(debugattack)
        cout << "$Putting an enemy group into atari." << endl;
      goto movexy;			/*    attack the group.*/
    }
  }
  if(upattern<BIGU) goto movepattern;
/*Try to harass the most vulnerable enemy group.*/
  while(ne)				/*For all remaining enemy groups */
  { if(attack(*es, &x, &y, 3))	/* if we can't see how to attack safely */
    { --ne; ++es; }			/*  go on to next enemy group*/
    else				/*   otherwise*/
    { if(debugattack)
        cout << "$Harassing the most vulnerable enemy group." << endl;
      goto movexy;			/*    attack it*/
    }
  }
#if 0 /*This seems to lose points.  Lets see how we do without it.*/
/*Move randomly adjacent to no stones or edges.*/
  for(x=1; x<thegame.boardSize-1; ++x)  for(y=1; y<thegame.boardSize-1; ++y)
    if( EMPTY==theboard[x][y] &&
	EMPTY==theboard[x][y-1] && EMPTY==theboard[x][y+1] &&
        EMPTY==theboard[x-1][y] && EMPTY==theboard[x+1][y] )
    { if(debugattack)
        cout << "$Moving randomly to fill space." << endl;
      goto movexy;
    }
#endif
/* If nothing else comes to mind, pass.*/
  thegame.kox= thegame.koy= (-1);
  movedone();
  thegame.wmove = "pass";
  if(thegame.qpa) return BOTHPASS;
  else
  { thegame.qpa= 1;
    return PASS;
  }

movexy:
  thegame.wx = x;
  thegame.wy = y;
  placestone(x, y, thegame.wallyColor);
  thegame.qpa= 0;
  movedone();
  
  // write vertex to string
  thegame.wmove = (char)lettercol(x);
  if ( y >= 9 ){
    if ( y >= 19 )
      thegame.wmove += "2";
    else
      thegame.wmove += "1";
    if ( y == 9 || y == 19 )
      thegame.wmove += "0";
    else
      thegame.wmove += ('1' + ( y % 10 ));
  }
  else
    thegame.wmove += ( '1' + y );

  return PLAY_MOVE;
}

int mymove()
{
  int move;
  move = enginemove();
  if ( (move == PASS) || (move == BOTHPASS) )
    cout << pname(thegame.wallyColor) << " passes." << endl;
  else {
    cout << pname(thegame.wallyColor) << " moves to ";
    cout.put(lettercol(thegame.wx));
    cout << thegame.wy+1 << "." << endl;
  }
  return move;
}

void opponentpass()
{
  movedone();
  thegame.kox = thegame.koy = (-1);
  lex = ley = (-1000);	/*'way off the board*/
  thegame.qpa= 1;
}
  
int enemymove()
/*Read and execute opponent's move.*/
{ int x, y;	/*coordinates of move*/
  int s;	/*return value of scanfcoo()*/

endrecurse:

  printgame();
  cout << "Game turn= " << thegame.tur << ".  ";
  cout << pname(thegame.pla) << " to play." << endl;
  cout << "Please input your move (" << pname(thegame.pla);
  cout << " #" << thegame.tur << "): ";
  s= scanfcoo(&x, &y);	/*Try to get a move.*/
  if(RESIGN==s)		/*If opponent resigns*/
   return RESIGN;
  if(PASS==s) {	/*If opponent passes*/
    movedone();
    thegame.kox= thegame.koy= (-1);
    lex = ley = (-1000);	/*'way off the board*/
    if(thegame.qpa) return BOTHPASS;
    else
    { thegame.qpa= 1;
      return PASS;
    }
  }
  if( 0==s )
  { cout << "Please try again.  Input your move as a letter for the column, \n";
    cout << " followed by a number for the row number.\n";
    goto endrecurse;
  }
  if( 0==onboard(x,y) )
  { cout << "Only moves which are on the board are legal.  Please try again.\n";
    goto endrecurse;
  }
  if( EMPTY!=theboard[x][y] )
  { cout << "You cannot stack a stone on another.  ";
pleasetryanothermove:
    cout << "Please try another move." << endl;
    goto endrecurse;
  }
  if(ko(x,y))
  { cout << "That move would violate the rule of ko.  ";
    goto pleasetryanothermove;
  }
  placestone(x, y, thegame.pla);
  thegame.qpa= 0;
    cout.put(lettercol(x));
    cout << y+1 << endl;

  movedone();
  return 0;
}


void cout_score()
{
  float finalScore;

  cout << "(the end of the game)\n";
  cout << "Komi is: " << setprecision(1) << thegame.komi << endl;
  cout << "Territory count is:\n";
  cout << "Black has " << thegame.territoryBlack << " territory." << endl;
  cout << "White has " << thegame.territoryWhite << " territory." << endl;
  finalScore = thegame.territoryBlack - thegame.territoryWhite - thegame.komi;
  if( 0 == finalScore )
    cout << "(a tie)\n";
  else if( finalScore > 0 )
    cout << "(Black wins.)\n";
  else
    cout << "(White wins.)\n";

}


int ist(int x, int y, int n, board scratch)

/*This is the recursive part of isterritory().*/

{
  bnassert(onboard(x,y));
  bnassert(EMPTY==theboard[x][y]);
  bnassert(0==scratch[x][y]);
  scratch[x][y]= 1;

  x= x+1;
  if(onboard(x,y))
  { if(EMPTY==theboard[x][y])
    { if(0==scratch[x][y] && 0==ist(x, y, n, scratch))
        return 0;
    }
    else if(n==theboard[x][y])
      return 0;
  }

  x= x-2;
  if(onboard(x,y))
  { if(EMPTY==theboard[x][y])
    { if(0==scratch[x][y] && 0==ist(x, y, n, scratch))
        return 0;
    }
    else if(n==theboard[x][y])
      return 0;
  }

  x= x+1;
  y= y+1;
  if(onboard(x,y))
  { if(EMPTY==theboard[x][y])
    { if(0==scratch[x][y] && 0==ist(x, y, n, scratch))
        return 0;
    }
    else if(n==theboard[x][y])
      return 0;
  }

  y= y-2;
  if(onboard(x,y))
  { if(EMPTY==theboard[x][y])
    { if(0==scratch[x][y] && 0==ist(x, y, n, scratch))
        return 0;
    }
    else if(n==theboard[x][y])
      return 0;
  }

  return 1;

}


int isterritory(int x, int y, int p)

/*
 Return TRUE if the EMPTY point x, y is to be counted as territory for
 player p.  This is only called at the end of the game, and I see no
 reason why it can't be slow.

 I don't understand the (Ing?) fractional point rules, so I don't implement them.
 An empty point counts as territory only if only one player has stones adjacent
 to the block of empty spaces to which it belongs.
*/
{
  board scratch;

  bnassert(onboard(x,y));
  bnassert(EMPTY==theboard[x][y]);

/*Clear scratch[].*/
  { register int x, y;
    for(x=0; x<thegame.boardSize; ++x)  for(y=0; y<thegame.boardSize; ++y)
      scratch[x][y]= 0;
  }

/*Decide recursively.*/
  return ist(x, y, nextp(p), scratch);

}

void score()
{
/*Count territories.*/
  estimatescore();
/*Output results.*/
  cout_score();
}

void estimatescore()
/*Compute and output the final game score.  Use Chinese rules, of course.*/
{
  register int x, y, p;
  int territory[NPLAYERS];	/*each player's territory*/

/*Count territories.*/
  for(p=0; p<NPLAYERS; ++p)
  { territory[p]= 0;
    for(x=0; x<thegame.boardSize; ++x)  for(y=0; y<thegame.boardSize; ++y)
      if(p==theboard[x][y]||EMPTY==theboard[x][y]&&isterritory(x,y,p))
	++territory[p];
  }
/* Save result in game class */
  thegame.territoryBlack = territory[BLACK];
  thegame.territoryWhite = territory[WHITE];
}

